/*

  Copyright (C) 2000 President and Fellows of Harvard College

  All rights reserved.

  NOTICE: This software is provided "as is", without any warranty,
  including any implied warranty for merchantability or fitness for a
  particular purpose.  Under no circumstances shall Harvard University
  or its faculty, staff, students or agents be liable for any use of,
  misuse of, or inability to use this software, including incidental
  and consequential damages.

  License is hereby given to use, modify, and redistribute this
  software, in whole or in part, for any commercial or non-commercial
  purpose, provided that the user agrees to the terms of this
  copyright notice, including disclaimer of warranty, and provided
  that this copyright notice, including disclaimer of warranty, is
  preserved in the source code and documentation of anything derived
  from this software.  Any redistributor of this software or anything
  derived from this software assumes responsibility for ensuring that
  any parties to whom such a redistribution is made are fully aware of
  the terms of this license and disclaimer.

  Author: Brad Karp, Harvard University EECS, May, 1999

*/

// GPSR for ns2 w/wireless extensions

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdarg.h>

#include "cmu-trace.h"
#include "random.h"
#include "mobilenode.h"
#include "locdbase.h"
#include "gpsr.h"

#define BEACON_RESCHED                                                     \
       beacon_timer_->resched(bint_ + Random::uniform(2*bdesync_* bint_) - \
			      bdesync_ * bint_)


inline double distance(double x1, double y1, double z1, double x2, double y2,
		       double z2)
{
  return sqrt((x1 - x2)*(x1 - x2) +
	      (y1 - y2)*(y1 - y2) +
	      (z1 - z2)*(z1 - z2));
}

inline double bearing(double x1, double y1, double x2, double y2)
{
  double brg;

  // XXX only deal with 2D for now
  // XXX check for (0, 0) args, a domain error
  brg = atan2(y2 - y1, x2 - x1);
  if (brg < 0)
    brg += 2*M_PI;
  return brg;
}

static int NeighbEntCmp(const void *a, const void *b)
{
  nsaddr_t ia = ((const NeighbEnt *) a)->dst;
  nsaddr_t ib = (*(const NeighbEnt **) b)->dst;
  if (ia > ib) return 1;
  if (ib > ia) return -1;
  return 0;
}

NeighbTable::NeighbTable(GPSR_Agent *mya)
{
  int i;

  nents = 0;
  maxents = 100;
  tab = new (NeighbEnt *)[100];
  a = mya;
  for (i = 0; i < 100; i++)
    tab[i] = new NeighbEnt(a);
}

NeighbTable::~NeighbTable()
{
  int i;

  // cancel timers
  for (i = 0; i < nents; i++) {
    tab[i]->dnt.force_cancel();
    tab[i]->ppt.force_cancel();
    delete tab[i];
  }
  delete[] tab;
}

NeighbTableIter NeighbTable::InitLoop() {
  return (NeighbTableIter) 0;
}

NeighbEnt *NeighbTable::NextLoop(NeighbTableIter *it) {
  if (((unsigned int) *it) >= (unsigned int) nents)
    return 0;

  return tab[(*it)++];
}

NeighbEnt *
NeighbTable::ent_findshortest(MobileNode *mn, double x, double y, double z)
{
  NeighbEnt *ne = 0;
  double shortest, t, myx, myy, myz;
  int i;

  mn->getLoc(&myx, &myy, &myz);
  shortest = distance(myx, myy, myz, x, y, z);
  for (i = 0; i < nents; i++)
    if ((t = distance(tab[i]->x, tab[i]->y, tab[i]->z, x, y, z)) < shortest) {
      shortest = t;
      ne = tab[i];
    }
  return ne;
}

NeighbEnt *
NeighbTable::ent_findcloser_onperi(MobileNode *mn, double x, double y,
				   double z, int *perihop)
{
  double mydist, t, myx, myy, myz;
  int i, j;

  mn->getLoc(&myx, &myy, &myz);
  mydist = distance(myx, myy, myz, x, y, z);
  for (i = 0; i < nents; i++)
    for (j = 0; j < tab[i]->perilen; j++)
      if ((t = distance(tab[i]->peri[j].x, tab[i]->peri[j].y,
			tab[i]->peri[j].z, x, y, z)) < mydist) {
	*perihop = j;
	return tab[i];
      }
  return 0;
}

int NeighbEnt::closer_pt(nsaddr_t myip, double myx, double myy, double myz,
			 double ptx, double pty,
			 nsaddr_t ptipa, nsaddr_t ptipb,
			 double dstx, double dsty,
			 double *closerx, double *closery)
{
  if ((min(dst, myip) == min(ptipa, ptipb)) &&
      (max(dst, myip) == max(ptipa, ptipb)))
    // this edge is the same edge where (ptx, pty) lies; nope
    return 0;
  if (!live)
    // this edge is not part of the planarized graph
    return 0;
  if (cross_segment(ptx, pty, dstx, dsty, myx, myy, x, y,
		    closerx, closery)) {
    if (distance(*closerx, *closery, 0.0, dstx, dsty, 0.0) <
	distance(ptx, pty, 0.0, dstx, dsty, 0.0)) {
      // edge has point closer than (ptx, pty)
      return 1;
    }
  }
  return 0;
}

NeighbEnt *
NeighbTable::ent_findcloser_edgept(MobileNode *mn, double ptx, double pty,
				   nsaddr_t ptipa, nsaddr_t ptipb,
				   double dstx, double dsty,
				   double *closerx, double *closery)
{
  NeighbTableIter ni;
  NeighbEnt *minne = NULL, *ne;
  double myx, myy, myz;

  mn->getLoc(&myx, &myy, &myz);
  ni = InitLoop();
  while ((ne = NextLoop(&ni))) {
    if (ne->closer_pt(mn->address(), myx, myy, myz, ptx, pty, ptipa, ptipb,
		      dstx, dsty, closerx, closery)) {
      // found an edge with a point closer than (ptx, pty)
      minne = ne;
      ptx = *closerx;
      pty = *closery;
      ptipa = mn->address();
      ptipb = ne->dst;
    }
  }
  return minne;
}

NeighbEnt *
NeighbTable::ent_findface(MobileNode *mn, double x, double y, double z, int p)
{
  double myx, myy, myz;
  double brg;

  // find bearing to dst
  mn->getLoc(&myx, &myy, &myz);
  brg = bearing(myx, myy, x, y);
  // find neighbor with greatest bearing <= brg
  return ent_next_ccw(brg, myx, myy, p);
}

NeighbEnt *
NeighbTable::ent_next_ccw(double basebrg, double x, double y, int p,
			  NeighbEnt *inne = 0)
{
  NeighbEnt *minne = NULL, *ne;
  NeighbTableIter ni;
  double brg, minbrg = 3*M_PI;

  ni = InitLoop();
  while ((ne = NextLoop(&ni))) {
    if (inne && (ne == inne))
      continue;
    if (p && !ne->live)
      continue;
    brg = bearing(x, y, ne->x, ne->y) - basebrg;
    if (brg < 0)
      brg += 2*M_PI;
    if (brg < 0)
      brg += 2*M_PI;
    if (brg < minbrg) {
      minbrg = brg;
      minne = ne;
    }
  }
  return minne;
}

NeighbEnt *
NeighbTable::ent_next_ccw(MobileNode *mn, NeighbEnt *inne, int p)
{
  double myx, myy, myz;
  double brg;
  NeighbEnt *ne;

  // find bearing from mn to (x, y, z)
  mn->getLoc(&myx, &myy, &myz);
  brg = bearing(myx, myy, inne->x, inne->y);
  ne = ent_next_ccw(brg, myx, myy, p, inne);
  if (!ne)
    return inne;
  else
    return ne;
}

NeighbEnt *NeighbTable::ent_finddst(nsaddr_t dst)
{
  NeighbEnt ne(NULL), **pne;

  ne.dst = dst;
  pne = ((NeighbEnt **) bsearch(&ne, tab, nents,
				sizeof(NeighbEnt *), NeighbEntCmp));
  if (pne)
    return *pne;
  else
    return NULL;
}

void NeighbTable::ent_delete(const NeighbEnt *ent)
{
  NeighbEnt **pne, *owslot;
  int i, j;

  if ((pne = (NeighbEnt **) bsearch(ent, tab, nents,
				    sizeof(NeighbEnt *), NeighbEntCmp))) {
    i = pne - tab;
    // make sure no timers scheduled for this neighbor
    (*pne)->dnt.force_cancel();
    (*pne)->ppt.force_cancel();
    // slide any subsequent table entries backward
    if (i < (nents - 1))
      owslot = tab[i];
    for (j = i; j < nents - 1; j++)
      tab[j] = tab[j+1];
    if (i < (nents - 1))
      tab[nents-1] = owslot;
    nents--;
  }
}

NeighbEnt *NeighbTable::ent_add(const NeighbEnt *ent)
{
  NeighbEnt **pne, *owslot;
  int i, j, r, l;

  if ((pne = (NeighbEnt **) bsearch(ent, tab, nents,
				    sizeof(NeighbEnt *), NeighbEntCmp))) {
    // already in table; overwrite
    // make sure there is no pending timer
    i = pne - tab;
    (*pne)->dnt.force_cancel();
    /* XXX overwriting table entry shouldn't affect when to probe this
       perimeter */
    // (*pne)->ppt.force_cancel();
    // careful not to overwrite timers!
    (*pne)->dst = ent->dst;
    (*pne)->x = ent->x;
    (*pne)->y = ent->y;
    (*pne)->z = ent->z;
    return *pne;
  }

  // may have to grow table
  if (nents == maxents) {
    NeighbEnt **tmp = tab;
    maxents *= 2;
    tab = new (NeighbEnt *)[maxents];
    bcopy(tmp, tab, nents*sizeof(NeighbEnt *));
    for (i = nents; i < maxents; i++)
      tab[i] = new NeighbEnt(a);
    delete[] tmp;
  }

  // binary search for insertion point
  if (nents == 0)
    i = 0;
  else {
    l = 0;
    r = nents - 1;
    while ((r - l) > 0) {
      if (ent->dst < tab[l + ((r-l) / 2)]->dst)
	r = l + ((r - l) / 2) - 1;
      else
	l += (r - l) / 2 + 1;
    }
    if (r < l)
      i = r+1;
    else
      // r == l
      if (ent->dst < tab[r]->dst)
	i = r;
      else
	i = r+1;
  }

  // slide subsequent entries forward
  if (i <= (nents - 1))
    owslot = tab[nents];
  j = nents-1;
  while (j >= i)
    tab[j+1] = tab[j--];
  // slam into table, without overwriting timers
  if (i <= (nents - 1))
    tab[i] = owslot;
  tab[i]->dst = ent->dst;
  tab[i]->x = ent->x;
  tab[i]->y = ent->y;
  tab[i]->z = ent->z;
  // invalidate the perimeter that may be cached by this neighbor entry
  tab[i]->perilen = 0;
  // XXX gross way to indicate entry is *new* entry, graph needs planarizing
  tab[i]->live = -1;
  nents++;
  return tab[i];
}

void NeighbEnt::planarize(NeighbTable *nt, int algo,
			  double x, double y, double z) {
  NeighbEnt *ne;
  NeighbTableIter ni;
  double uvdist, canddist, midx, midy;

  uvdist = distance(x, y, z, this->x, this->y, this->z);
  switch(algo) {
  case PLANARIZE_RNG:
    break;
  case PLANARIZE_GABRIEL:
    // find midpt of segment me (u) <-> this (v)
    midx = (x + this->x) / 2.0;
    midy = (y + this->y) / 2.0;
    uvdist /= 2.0;
    break;
  default:
    fprintf(stderr, "Unknown graph planarization algorithm %d\n", algo);
    abort();
    break;
  }
  ni = nt->InitLoop();
  while ((ne = nt->NextLoop(&ni))) {
    if (ne == this)
      // w and v identical node--w not a witness
      continue;
    switch(algo) {
    case PLANARIZE_RNG:
      // find max dist. from me (u) to ne (w) vs. this (v) to ne (w)
      canddist = max(distance(x, y, z, ne->x, ne->y, ne->z),
		     distance(this->x, this->y, this->z, ne->x, ne->y, ne->z));
      // is max < dist from me (u) to this (v)?
      if (canddist < uvdist) {
	this->live = 0;
	return;
      }
      break;
    case PLANARIZE_GABRIEL:
      // is ne (w) inside circle of radius uvdist?
      if (distance(midx, midy, 0.0, ne->x, ne->y, 0.0) < uvdist) {
	this->live = 0;
	return;
      }
      break;
    default:
      break;
    }
  }
  this->live = 1;
}

void NeighbTable::planarize(int algo, double x, double y, double z)
{
  NeighbEnt *ne;
  NeighbTableIter ni;

  ni = InitLoop();
  while ((ne = NextLoop(&ni))) {
    ne->planarize(this, algo, x, y, z);
  }
}

int hdr_gpsr::offset_;

class GPSRHeaderClass : public PacketHeaderClass {
public: 
  GPSRHeaderClass() : PacketHeaderClass("PacketHeader/GPSR", sizeof(hdr_gpsr))
  {
    bind_offset(&hdr_gpsr::offset_);
  }
} class_gpsrhdr;

static class GPSRClass:public TclClass
{
public:
  GPSRClass():TclClass ("Agent/GPSR")
  {
  }
  TclObject *create (int, const char *const *)
  {
    return (new GPSR_Agent ());
  }
} class_gpsr;

GPSR_Agent::GPSR_Agent(void) : Agent(PT_GPSR), beacon_timer_(0),
  lastperi_timer_(0), use_mac_(0),
  use_peri_(0), verbose_(1), drop_debug_(0), peri_proact_(1),
  use_implicit_beacon_(0), use_planar_(0), use_loop_detect_(0),
  use_timed_plnrz_(0), bint_(GPSR_ALIVE_INT), bdesync_(GPSR_ALIVE_DESYNC),
  bexp_(GPSR_ALIVE_EXP), pint_(GPSR_PPROBE_INT), pdesync_(GPSR_PPROBE_DESYNC),
  lpexp_(GPSR_PPROBE_EXP), ldb_(0), mn_(0), ifq_(0)
{
  ntab_ = new NeighbTable(this);

  bind_time("bint_", &bint_);
  bind_time("bexp_", &bexp_);
  bind_time("bdesync_", &bdesync_);
  bind_time("pint_", &pint_);
  bind_time("pdesync_", &pdesync_);
  bind_time("lpexp_", &lpexp_);
  bind("verbose_", &verbose_);
  bind("drop_debug_", &drop_debug_);
  bind("peri_proact_", &peri_proact_);
  bind("use_mac_", &use_mac_);
  bind("use_peri_", &use_peri_);
  bind("use_implicit_beacon_", &use_implicit_beacon_);
  bind("use_planar_", &use_planar_);
  bind("use_loop_detect_", &use_loop_detect_);
  bind("use_timed_plnrz_", &use_timed_plnrz_);
  bind("off_gpsr_", &off_gpsr_);
}

void GPSR_Agent::trace(char *fmt,...)
{
  va_list ap;

  if (!tracetarget)
    return;

  va_start(ap, fmt);
  vsprintf(tracetarget->buffer(), fmt, ap);
  tracetarget->dump();
  va_end(ap);
}

void GPSR_Agent::tracepkt(Packet *p, double now, int me, const char *type)
{
  char buf[1024];

  hdr_gpsr *gpsrh = hdr_gpsr::access(p);

  snprintf (buf, 1024, "V%s %.5f _%d_:", type, now, me);

  if (gpsrh->mode_ == GPSRH_BEACON) {
    snprintf (buf, 1024, "%s (%f,%f,%f)", buf, gpsrh->hops_[0].x,
	      gpsrh->hops_[0].y, gpsrh->hops_[0].z);
    if (verbose_)
      trace("%s", buf);
  }
}

// don't drop or modify the packet--it's not a copy!
void GPSR_Agent::tap(const Packet *p)
{
  hdr_cmn *hdrc = HDR_CMN(p);
  hdr_gpsr *gpsrh = (hdr_gpsr *) p->access(off_gpsr_);

  /* ignore non-IP packets.
     ignore beacons; we process those on regular receive.
     assumes the MAC tap includes all unicast packets bound for us.
     process those here, and avoid calls to beacon_proc() elsewhere. */
  if (use_implicit_beacon_ &&
      (hdrc->addr_type_ == AF_INET) &&
      (hdrc->next_hop_ != (nsaddr_t) IP_BROADCAST)) {
    // snoop it as proof of its sender's existence.
    switch (gpsrh->mode_) {
    case GPSRH_DATA_GREEDY:
      // prev hop position lives in hops_[0]
      beacon_proc(gpsrh->hops_[0].ip, gpsrh->hops_[0].x, gpsrh->hops_[0].y,
		  gpsrh->hops_[0].z);
      break;
    case GPSRH_PPROBE:
      // prev hop position lives in hops_[nhops_-1]
      beacon_proc(gpsrh->hops_[gpsrh->nhops_-1].ip,
		  gpsrh->hops_[gpsrh->nhops_-1].x,
		  gpsrh->hops_[gpsrh->nhops_-1].y,
		  gpsrh->hops_[gpsrh->nhops_-1].z);
      break;
    case GPSRH_DATA_PERI:
      // XXX was hops_[gpsrh->currhop_-1]
      // prev hop position lives in hops_[0]
      beacon_proc(gpsrh->hops_[0].ip,
		  gpsrh->hops_[0].x,
		  gpsrh->hops_[0].y,
		  gpsrh->hops_[0].z);
      break;
    default:
      fprintf(stderr, "Yow! tap got packet of unk type %d!\n", gpsrh->mode_);
      abort();
      break;
    }
  }
}

void GPSR_Agent::lost_link(Packet *p)
{
  hdr_cmn *hdrc = HDR_CMN(p);
  hdr_gpsr *gpsrh = hdr_gpsr::access(p);
  NeighbEnt *ne;

  if (use_mac_ == 0) {
    drop(p, DROP_RTR_MAC_CALLBACK);
    return;
  }

  //DEBUG
  //printf("(%d)..Lost link..\n",myaddr_);
  if (verbose_ && hdrc->addr_type_ == AF_INET)
    trace ("VLL %.8f %d->%d lost at %d",
	   Scheduler::instance().clock(),
	   ((hdr_ip *) p->access(off_ip_))->src_,
	   ((hdr_ip *) p->access(off_ip_))->dst_,
	   mn_->address());

  if (hdrc->addr_type_ == AF_INET) {
    ne = ntab_->ent_finddst(hdrc->next_hop_);
    if (verbose_)
      trace ("VLP %.5f %d:%d->%d:%d lost at %d [hop %d]",
	     Scheduler::instance().clock(),
	     ((hdr_ip *) p->access(off_ip_))->src_,
	     ((hdr_ip *) p->access(off_ip_))->sport_,
	     ((hdr_ip *) p->access(off_ip_))->dst_,
	     ((hdr_ip *) p->access(off_ip_))->dport_,
	     mn_->address(), hdrc->next_hop_);

    if (ne) {
      ne->dnt.force_cancel();
      deadneighb_callback(ne);
    }
  }

  // grab packets in the ifq bound for the same next hop
  Packet *r, *rh, *rt;
  rh = rt = p;
  rh->next_ = 0;
  while ((r = ifq_->filter((nsaddr_t) hdrc->next_hop_))) {
    rt->next_ = r;
    r->next_ = 0;
    rt = r;
  }

  // retarget all packets we got to a new next hop
  while (rh) {
    rt = rh;
    rh = rh->next_;

    hdrc = HDR_CMN(rt);
    gpsrh = hdr_gpsr::access(rt);

    if (hdrc->addr_type_ != AF_INET) {
      drop(rt, DROP_RTR_MAC_CALLBACK);
      continue;
    }

    // for GPSR perimeter probes, chop off our own perimeter entry before
    // passing the probe back into the agent for reforwarding
    if (((hdr_ip *) rt->access(off_ip_))->dport_ == GPSR_PORT) {
      if (gpsrh->mode_ == GPSRH_PPROBE) {
	if (gpsrh->nhops_ == 1) {
	  /* we originated it. the neighbor is gone, according to the MAC
	     layer. drop the probe--it was *only* meant for that neighbor. */
	  drop(rt, DROP_RTR_NEXT_SRCRT_HOP);
	  continue;
	}
	/* we were forwarding the probe, so instead try to recover by
	   forwarding it to a remaining appropriate next hop */
	gpsrh->nhops_--;
	hdrc->size() = gpsrh->size() + IP_HDR_LEN;
	periIn(rt, gpsrh, GPSR_PPROBE_RTX);
      }
    }
    else {
      switch (gpsrh->mode_) {
      case GPSRH_DATA_GREEDY:
	// give the packet another chance--exercise greedy's good recovery
	forwardPacket(rt);
	break;
      case GPSRH_DATA_PERI:
	if (use_planar_)
	  // not src-routed; give it another chance via another neighbor
	  forwardPacket(rt, 1);
	else
	  // punt the packet; its chosen src-routed next hop is gone
	  drop(rt, DROP_RTR_NEXT_SRCRT_HOP);
	break;
      default:
	fprintf(stderr,
		"yow! non-data packet for non-GPSR port bounced by MAC!\n");
	abort();
	break;
      }
    }
  }
}

static void mac_callback(Packet * p, void *arg)
{
  ((GPSR_Agent *) arg)->lost_link(p);
}

// construct and return an alive beacon packet
Packet *GPSR_Agent::makeAlive(void)
{
  Packet *p = allocpkt();
  hdr_ip *iph = (hdr_ip *) p->access (off_ip_);
  hdr_cmn *hdrc = HDR_CMN(p);
  hdr_gpsr *gpsrh = hdr_gpsr::access(p);

  // set up header
  hdrc->next_hop_ = IP_BROADCAST;
  hdrc->addr_type_ = AF_INET;
  hdrc->ptype_ = PT_GPSR;
  iph->dst_ = IP_BROADCAST << Address::instance().nodeshift();
  iph->dport_ = GPSR_PORT;
  gpsrh->mode_ = GPSRH_BEACON;
  gpsrh->nhops_ = 1;
  mn_->getLoc(&gpsrh->hops_[0].x, &gpsrh->hops_[0].y, &gpsrh->hops_[0].z);
  // simulated size: hdr_gpsr, IP header
  hdrc->size() = gpsrh->size() + IP_HDR_LEN;
  return p;
}

void GPSR_Agent::planar_callback(void)
{
  // re-planarize graph
  if (use_planar_) {
    double myx, myy, myz;

    mn_->getLoc(&myx, &myy, &myz);
    ntab_->planarize(PLANARIZE_RNG, myx, myy, myz);
  }
  // reschedule us
  // XXX should make interval tunable!!!
  planar_timer_->resched(1.0);
}

void GPSR_Agent::lastperi_callback(void)
{
  NeighbEnt *ne;
  NeighbTableIter ni;

  // don't probe perimeters proactively anymore
  peri_proact_ = 0;
  // cancel all perimeter probe timers
  ni = ntab_->InitLoop();
  while ((ne = ntab_->NextLoop(&ni)))
    ne->ppt.force_cancel();
}

void GPSR_Agent::beacon_callback(void)
{
  Scheduler &s = Scheduler::instance();

  Packet *p = makeAlive();

  // schedule the transmission of this beacon
  if (p) {
    assert (!HDR_CMN(p)->xmit_failure_);
    s.schedule (target_, p, 0);
  }
  // schedule the next beacon generation event
  BEACON_RESCHED;
}

void GPSR_Agent::deadneighb_callback(NeighbEnt *ne)
{
  Scheduler &s = Scheduler::instance();
  double now = s.clock ();

  if (verbose_)
    trace ("VTO %.5f _%d_ %d->%d", now, mn_->address(), mn_->address(),
	   ne->dst);
  // remove the neighbor entry from the table!
  ntab_->ent_delete(ne);
  // need to re-planarize, if option dictates
  if (use_planar_) {
    double myx, myy, myz;

    mn_->getLoc(&myx, &myy, &myz);
    ntab_->planarize(PLANARIZE_RNG, myx, myy, myz);
  }
}

void GPSR_Agent::periprobe_callback(NeighbEnt *ne)
{
  Packet *p = allocpkt();
  hdr_ip *iph = (hdr_ip *) p->access (off_ip_);
  hdr_gpsr *gpsrh = hdr_gpsr::access(p);
  hdr_cmn *ch = HDR_CMN(p);

  ch->next_hop_ = ne->dst;
  ch->addr_type_ = AF_INET;
  ch->ptype_ = PT_GPSR;
  iph->dst_ = Address::instance().create_ipaddr(ne->dst, GPSR_PORT);
  iph->dport_ = GPSR_PORT;
  iph->ttl() = 128;
  gpsrh->hops_[0].ip = Address::instance().get_nodeaddr(addr_);
  gpsrh->nhops_ = 1;
  gpsrh->mode_ = GPSRH_PPROBE;
  // simulated size: hdr_gpsr and hdr_ip
  ch->size() = gpsrh->size() + IP_HDR_LEN;
  mn_->getLoc(&gpsrh->hops_[0].x, &gpsrh->hops_[0].y, &gpsrh->hops_[0].z);

  // schedule probe transmission
  ch->xmit_failure_ = mac_callback;
  ch->xmit_failure_data_ = this;
  Scheduler::instance().schedule(target_, p, 0);
  if (use_implicit_beacon_)
    BEACON_RESCHED;

  // schedule next probe timer
  ne->ppt.resched(pint_ +
		  Random::uniform(2 * pdesync_ * pint_) - pdesync_ * pint_);
}

void GPSR_Agent::beacon_proc(int src, double x, double y, double z)
{
  NeighbEnt *ne;
  NeighbEnt nne(this);

  nne.dst = src;
  nne.x = x;
  nne.y = y;
  nne.z = z;
  ne = ntab_->ent_add(&nne);
  if (ne->live == -1) {
    // entry wasn't in table before. need to planarize, if option dictates.
    ne->live = 1;
    if (use_planar_) {
      double myx, myy, myz;

      mn_->getLoc(&myx, &myy, &myz);
      ntab_->planarize(PLANARIZE_RNG, myx, myy, myz);
    }
  }
  ne->dnt.resched(bexp_);
  if (use_peri_ && (ne->ppt.status() == TIMER_IDLE))
    // no perimeter probe is pending; launch one
    if (peri_proact_)
      ne->ppt.sched(pint_ +
		    Random::uniform(2 * pdesync_ * pint_) - pdesync_ * pint_);
}

void GPSR_Agent::beaconIn(Packet *p, hdr_gpsr *gpsrh)
{
  hdr_ip *hdrip = (hdr_ip *) p->access(off_ip_);
  int src = Address::instance().get_nodeaddr(hdrip->src_);

  beacon_proc(src, gpsrh->hops_[0].x, gpsrh->hops_[0].y, gpsrh->hops_[0].z);
  // we consumed the packet; free it!
  Packet::free(p);
}

inline int cross_segment(double x1, double y1, double x2, double y2,
			 double x3, double y3, double x4, double y4,
			 double *xi = 0, double *yi = 0)
{
  double dy[2], dx[2], m[2], b[2];
  double xint, yint;

  dy[0] = y2 - y1;
  dx[0] = x2 - x1;
  dy[1] = y4 - y3;
  dx[1] = x4 - x3;
  m[0] = dy[0] / dx[0];
  m[1] = dy[1] / dx[1];
  b[0] = y1 - m[0] * x1;
  b[1] = y3 - m[1] * x3;
  if (m[0] != m[1]) {
    // slopes not equal, compute intercept
    xint = (b[0] - b[1]) / (m[1] - m[0]);
    yint = m[1] * xint + b[1];
    // is intercept in both line segments?
    if ((xint <= max(x1, x2)) && (xint >= min(x1, x2)) &&
	(yint <= max(y1, y2)) && (yint >= min(y1, y2)) &&
	(xint <= max(x3, x4)) && (xint >= min(x3, x4)) &&
	(yint <= max(y3, y4)) && (yint >= min(y3, y4))) {
      if (xi && yi) {
	*xi = xint;
	*yi = yint;
      }
      return 1;
    }
  }
  return 0;
}

int GPSR_Agent::crosses(NeighbEnt *ne, hdr_gpsr *gpsrh)
{
  int i;

  // check all neighboring hops in perimeter thus far (through self)
  for (i = 0; i < (gpsrh->nhops_ - 1); i++) {
    if ((gpsrh->hops_[i].ip != ne->dst) &&
	(gpsrh->hops_[i+1].ip != ne->dst) &&
	(gpsrh->hops_[i].ip != gpsrh->hops_[gpsrh->nhops_-1].ip) &&
	(gpsrh->hops_[i+1].ip != gpsrh->hops_[gpsrh->nhops_-1].ip) &&
	cross_segment(gpsrh->hops_[i].x, gpsrh->hops_[i].y,
		      gpsrh->hops_[i+1].x, gpsrh->hops_[i+1].y,
		      gpsrh->hops_[gpsrh->nhops_-1].x,
		      gpsrh->hops_[gpsrh->nhops_-1].y,
		      ne->x, ne->y))
      return 1;
  }
  return 0;
}

void GPSR_Agent::periIn(Packet *p, hdr_gpsr *gpsrh, int rtxflag = 0)
{
  double myx, myy, myz;
  NeighbEnt *ne, *inne;

  // update neighbor record for previous hop
  
  // did I originate it?
  if (gpsrh->hops_[0].ip == Address::instance().get_nodeaddr(addr_)) {
    // cache the perimeter
    ne = ntab_->ent_finddst(gpsrh->hops_[1].ip);
    if (!ne) {
      // apparently, neighbor we launched probe via is now gone
      Packet::free(p);
      return;
    }
#ifdef HDR_GPSR_DYNAMIC
    if (ne->peri && (ne->maxlen < gpsrh->maxhops_)) {
      // need to allocate more PeriEnt slots in ne
      delete[] ne->peri;
      ne->maxlen = ne->perilen = 0;
      ne->peri = NULL;
    }
#endif
    if (!ne->peri) {
#ifdef HDR_GPSR_DYNAMIC
      ne->peri = new (struct PeriEnt)[gpsrh->maxhops_];
      ne->maxlen = gpsrh->maxhops_;
#else
      ne->peri = new (struct PeriEnt)[MAX_PERI_HOPS_STATIC];
      ne->maxlen = MAX_PERI_HOPS_STATIC;
#endif
    }
    bcopy(&gpsrh->hops_[1], ne->peri,
	  (gpsrh->nhops_ - 1) * sizeof(struct PeriEnt));
    ne->perilen = gpsrh->nhops_ - 1;
    /* no timer work to do--perimeter probe timer is governed by
       beacons/absence of beacons from a neighbor */
    // we consumed the packet; free it!
    Packet::free(p);
    return;
  }
  // add self to GPSR header perimeter
  mn_->getLoc(&myx, &myy, &myz);
  gpsrh->add_hop(Address::instance().get_nodeaddr(addr_), myx, myy, myz);
  // compute candidate next hop: sweep ccw about self from ingress hop
  ne = inne = ntab_->ent_finddst(gpsrh->hops_[gpsrh->nhops_-2].ip);
  /* in theory, a perimeter probe received from an unknown neighbor should
     serve as a beacon from that neighbor... */
  /* BUT, don't add the previous hop more than once when we retransmit a
     peri probe--the prev hop information is stale in that case */
  if (!rtxflag && (ne == NULL)) {
    NeighbEnt nne(this);

    nne.dst = gpsrh->hops_[gpsrh->nhops_-2].ip;
    nne.x = gpsrh->hops_[gpsrh->nhops_-2].x;
    nne.y = gpsrh->hops_[gpsrh->nhops_-2].y;
    nne.z = gpsrh->hops_[gpsrh->nhops_-2].z;
    inne = ne = ntab_->ent_add(&nne);
    ne->dnt.sched(bexp_);
    // no perimeter probe is pending; launch one
    if (peri_proact_)
      ne->ppt.sched(pint_ +
		    Random::uniform(2 * pdesync_ * pint_) - pdesync_ * pint_);
  }
  else if (ne == NULL) {
    /* we're trying to retransmit a peri probe, but the ingress hop is gone.
       drop it. */
    drop(p, DROP_RTR_MAC_CALLBACK);
    return;
  }
  while ((ne = ntab_->ent_next_ccw(mn_, ne, use_planar_)) != inne)
    // verify no crossing
    if (!crosses(ne, gpsrh))
      break;
  // forward probe to ne
  hdr_cmn *cmh = (hdr_cmn *) p->access(off_cmn_);
  hdr_ip *iph = (hdr_ip *) p->access(off_ip_);
  cmh->size() += sizeof(struct PeriEnt);
  cmh->addr_type_ = AF_INET;
  cmh->xmit_failure_ = mac_callback;
  cmh->xmit_failure_data_ = this;
  cmh->next_hop_ = ne->dst;
  cmh->direction_ = -1;
  iph->dst_ = Address::instance().create_ipaddr(ne->dst, GPSR_PORT);
  target_->recv(p, (Handler *)0);
  if (use_implicit_beacon_)
    BEACON_RESCHED;
}

void GPSR_Agent::forwardPacket(Packet *p, int rtxflag = 0)
{
  hdr_ip *iph = (hdr_ip *) p->access(off_ip_);
  hdr_cmn *cmh = (hdr_cmn *) p->access(off_cmn_);
  hdr_gpsr *gpsrh = (hdr_gpsr *) p->access(off_gpsr_);
  NeighbEnt *ne, *logne;
  Scheduler &s = Scheduler::instance();
  double now = s.clock();

  switch(gpsrh->mode_) {
  case GPSRH_DATA_GREEDY:
    ne = ntab_->ent_findshortest(mn_, iph->dx_, iph->dy_, iph->dz_);

    if (ne)
      cmh->next_hop_ = ne->dst;
    else {
      int closer;

      if (use_peri_ && use_planar_) {
	// no proactive probes, so no peri_proact_ to worry about
	ne = ntab_->ent_findface(mn_, iph->dx_, iph->dy_, iph->dz_,
				 use_planar_);
	if (!ne) {
	  // no face toward the destination
	  drop(p, DROP_RTR_NO_ROUTE);
	  return;
	}
	// put packet in peri data mode, forward
	gpsrh->mode_ = GPSRH_DATA_PERI;
        // mark point of entry into peri data mode
	mn_->getLoc(&gpsrh->peript_.x, &gpsrh->peript_.y, &gpsrh->peript_.z);
	// mark ips of edge endpoints
	gpsrh->periptip_[0] = ne->dst; // prev edge on peri
	gpsrh->periptip_[1] = mn_->address(); // myself
	gpsrh->periptip_[2] = ne->dst; // next edge on peri
	/* N.B. first dst hop is hops_[1]
	   (leave room for hop-by-hop ip, position in hops_[0])!! */
	gpsrh->nhops_ = 1;
	gpsrh->currhop_ = 1;
	gpsrh->add_hop(mn_->address(), gpsrh->peript_.x, gpsrh->peript_.y,
		       gpsrh->peript_.z);
	cmh->next_hop_ = ne->dst;
	// add 20 bytes (pos, 3 X edge endpt ips) of entry point to peri mode
	/* subtract 12 bytes: if use_implicit_beacon, hop-by-hop pos/ip
	   already accounted for in size; if not, don't count hops_[0]! */
	cmh->size() += gpsrh->size() + 20 - 12;
      }
      else if (use_peri_ && peri_proact_) {
	// record we had a data packet that needed a perimeter
	if (lastperi_timer_)
	  lastperi_timer_->resched(lpexp_);
	// but wait! maybe we can get closer along a perimeter.
	ne = ntab_->ent_findcloser_onperi(mn_, iph->dx_, iph->dy_, iph->dz_,
					  &closer);
	if (!ne) {
	  // we're well and truly hung; nothing closer on a peri, either
	  if (drop_debug_ && (cmh->opt_num_forwards_ != 16777215)) {
	    NeighbTableIter ni;
	    ni = ntab_->InitLoop();
	    while ((logne = ntab_->NextLoop(&ni))) {
	      trace("VPER _%d_ (%.5f, %.5f):",
		    logne->dst, logne->x, logne->y);
	      for (int j = 0; j < logne->perilen; j++) {
		trace("VPER\t\t_%d_ (%.5f, %.5f)",
			logne->peri[j].ip, logne->peri[j].x, logne->peri[j].y);
	      }
	    }
	  }
	  drop(p, DROP_RTR_NO_ROUTE);
	  return;
	}
	else {
	  // put packet in peri mode
	  gpsrh->mode_ = GPSRH_DATA_PERI;
	  /* N.B. first dst hop is hops_[1]
	     (leave room for hop-by-hop ip, position in hops_[0])!! */
	  gpsrh->nhops_ = 1;
	  gpsrh->currhop_ = 1;
	  // add gpsrh with *shorter* path to closer node
	  int i;
	  if (closer < (ne->perilen / 2)) {
	    for (i = 0; i <= closer; i++)
	      gpsrh->add_hop(ne->peri[i].ip, ne->peri[i].x, ne->peri[i].y,
			     ne->peri[i].z);
	  }
	  else {
	    for (i = ne->perilen-1; i >= closer; i--)
	      gpsrh->add_hop(ne->peri[i].ip, ne->peri[i].x, ne->peri[i].y,
			     ne->peri[i].z);
	  }
	  cmh->next_hop_ = gpsrh->hops_[1].ip;
	  /* subtract 12 bytes: if use_implicit_beacon, hop-by-hop pos/ip
	     already accounted for in size; if not, don't count hops_[0]! */
	  cmh->size() += gpsrh->size() - 12;
	}
      }
      else {
	// no closer neighbor! unforwardable; drop it.
	/* someday, may want to queue up packets for currently unforwardable
	   destinations */
	// record we had a data packet that needed a perimeter
	if (lastperi_timer_)
	  lastperi_timer_->resched(lpexp_);
	// we could have used a perimeter here--turn them on
	peri_proact_ = 1;
	drop(p, DROP_RTR_NO_ROUTE);
	return;
      }
    }
    break;
  case GPSRH_DATA_PERI:
    if (use_peri_) {
      if (use_planar_) {
	double myx, myy, myz, closerx, closery;

	// non-source-routed perimeter forwarding rule
	/* to resume greedy forwarding, this *node* must be closer than
	   the point where the packet entered peri mode. */
	mn_->getLoc(&myx, &myy, &myz);
	if (distance(myx, myy, myz, iph->dx_, iph->dy_, iph->dz_) <
	    distance(gpsrh->peript_.x, gpsrh->peript_.y, gpsrh->peript_.z,
		     iph->dx_, iph->dy_, iph->dz_)) {
	  gpsrh->mode_ = GPSRH_DATA_GREEDY;
	  /* always add back (- - is +) 12 bytes: if use_implicit_beacon_,
	     src added 12 to size, don't re-add hops_[0]; otherwise,
	     still don't want to count hops_[0]. */
	  cmh->size() -= gpsrh->size() + 20 - 12;
	  gpsrh->currhop_ = 0;
	  gpsrh->nhops_ = 0;
	  // recursive, but must call target_->recv in callee frame
	  forwardPacket(p);
	  return;
	}
	// forward along current face, or change faces where appropriate
	/* don't choose *any* edge--only consider edges on the
	   face we're forwarding on at the moment. */
	ne = ntab_->ent_finddst(gpsrh->hops_[gpsrh->nhops_-1].ip);
	if (ne) {
	  ne = ntab_->ent_next_ccw(mn_, ne, use_planar_);
	  /* drop if we've looped on this perimeter:
	     are about to revisit the first edge we took on it */
	  if ((gpsrh->periptip_[1] == mn_->address()) &&
	      (gpsrh->periptip_[2] == ne->dst)) {
	    drop(p, DROP_RTR_NO_ROUTE);
	    return;
	  }
	  if (use_loop_detect_) {
	    /* if we've looped at a point *other* than the first hop on
	       the perimeter, clear the packet's state (other than peript,
	       which is still the "point to beat"), and forward it along the
	       appropriate face. */
	    int j;
	    for (j = 1; j < gpsrh->nhops_-1; j++) {
	      if ((gpsrh->hops_[j].ip == mn_->address()) &&
		  (gpsrh->hops_[j+1].ip == ne->dst)) {
		ne = ntab_->ent_findface(mn_, iph->dx_, iph->dy_, iph->dz_,
					 use_planar_);
		if (!ne) {
		  // no face toward the destination
		  drop(p, DROP_RTR_NO_ROUTE);
		  return;
		}
		cmh->size() -= gpsrh->size() + 20 - 12;
		gpsrh->periptip_[0] = ne->dst; // prev edge on peri
		gpsrh->periptip_[1] = mn_->address(); // myself
		gpsrh->periptip_[2] = ne->dst; // next edge on peri
		gpsrh->nhops_ = 1;
		gpsrh->currhop_ = 1;
		gpsrh->add_hop(mn_->address(), myx, myy, myz);
		cmh->next_hop_ = ne->dst;
		/* add 20 bytes (pos, 3 X edge endpt ips) of entry point to
		   peri mode */
		/* subtract 12 bytes: if use_implicit_beacon, hop-by-hop pos/ip
		   already accounted for in size; if not, don't count
		   hops_[0]! */
		cmh->size() += gpsrh->size() + 20 - 12;
		// XXX gotta fix this, gross...
		goto finish_pkt;
	      }
	    }
	  }
	  // does the candidate next edge have a closer pt?
	  if (ne->closer_pt(mn_->address(), myx, myy, myz,
			    gpsrh->peript_.x, gpsrh->peript_.y,
			    gpsrh->periptip_[1], gpsrh->periptip_[0],
			    iph->dx_, iph->dy_, &closerx, &closery)) {
	    /* yes. choose a new next hop on the peri pierced by the line
	       to the destination. */
	    /* several neighboring edges may be cut by line to destination;
	       choose that cut at the point closest to destination */
	    while (ne->closer_pt(mn_->address(), myx, myy, myz,
				 gpsrh->peript_.x, gpsrh->peript_.y,
				 gpsrh->periptip_[1], gpsrh->periptip_[0],
				 iph->dx_, iph->dy_, &closerx, &closery)) {
	      // fake that ingress edge was edge from ne
	      if (use_loop_detect_) {
		// clear all old hops, record new first one
		cmh->size() -= gpsrh->size() + 20 - 12;
		gpsrh->nhops_ = 1;
		gpsrh->currhop_ = 1;
		gpsrh->add_hop(ne->dst, ne->x, ne->y, ne->z);
		cmh->size() += gpsrh->size() + 20 - 12;
	      }
	      else {
		// re-use single-hop history
		gpsrh->hops_[gpsrh->nhops_-1].ip = ne->dst;
		gpsrh->hops_[gpsrh->nhops_-1].x = ne->x;
		gpsrh->hops_[gpsrh->nhops_-1].y = ne->y;
		gpsrh->hops_[gpsrh->nhops_-1].z = ne->z;
	      }
	      // record closest point on edge to ne
	      gpsrh->peript_.x = closerx;
	      gpsrh->peript_.y = closery;
	      gpsrh->peript_.z = 0.0;
	      ne = ntab_->ent_next_ccw(mn_, ne, use_planar_);
	    }
	    // record edge endpt ips
	    gpsrh->periptip_[0] = ne->dst; // prev hop
	    gpsrh->periptip_[1] = mn_->address(); // self
	    gpsrh->periptip_[2] = ne->dst; // next hop
	    forwardPacket(p);
	    return;
	  }
	}
	// forward to next ccw neighbor from ingress edge
	/* in theory, a data peri packet received from an unknown neighbor
	   should serve as a beacon from that neighbor... */
	/* BUT, don't add the previous hop more than once when we retransmit a
	   packet--the prev hop information is stale in that case */
	if (!rtxflag && (ne == NULL)) {
	  NeighbEnt nne(this);

	  nne.dst = gpsrh->hops_[gpsrh->nhops_-1].ip;
	  nne.x = gpsrh->hops_[gpsrh->nhops_-1].x;
	  nne.y = gpsrh->hops_[gpsrh->nhops_-1].y;
	  nne.z = gpsrh->hops_[gpsrh->nhops_-1].z;
	  ne = ntab_->ent_add(&nne);
	  ne->dnt.sched(bexp_);
	  // now that we have state for the ingress edge, try again...
	  forwardPacket(p);
	  return;
	}
	else if (ne == NULL) {
	  // XXX might we now be able to forward anyway?? know loc of prev hop.
	  /* we're trying to retransmit a packet, but the ingress hop is
	     gone. drop it. */
	  drop(p, DROP_RTR_MAC_CALLBACK);
	  return;
	}
	cmh->next_hop_ = ne->dst;
	if (use_loop_detect_) {
	  gpsrh->add_hop(mn_->address(), myx, myy, myz);
	  cmh->size() += 12;
	}
	else {
	  gpsrh->hops_[gpsrh->nhops_-1].ip = mn_->address();
	  gpsrh->hops_[gpsrh->nhops_-1].x = myx;
	  gpsrh->hops_[gpsrh->nhops_-1].y = myy;
	  gpsrh->hops_[gpsrh->nhops_-1].z = myz;
	}
      }
      else {
	// am I the right waypoint?
	if (gpsrh->hops_[gpsrh->currhop_].ip == mn_->address()) {
	  // am I the final waypoint?
	  if (gpsrh->currhop_ == (gpsrh->nhops_-1)) {
	    // yes! return packet to greedy mode
	    gpsrh->mode_ = GPSRH_DATA_GREEDY;
	    cmh->size() -= gpsrh->size() - 12;
	    gpsrh->currhop_ = 0;
	    gpsrh->nhops_ = 0;
	    // recursive, but must either drop or target_->recv in callee frame
	    forwardPacket(p);
	    return;
	  }
	  else {
	    // forward using source route...
	    gpsrh->currhop_++;
	    cmh->next_hop_ = gpsrh->hops_[gpsrh->currhop_].ip;
	  }
	}
	else {
	  // topology must have changed; I'm not the right hop
	  drop(p, DROP_RTR_NO_ROUTE);
	  return;
	}
      }
    }
    else {
      fprintf(stderr,
	      "yow! got peri mode packet when not using perimeters!\n");
      abort();
    }
    break;
  default:
    fprintf(stderr, "yow! got non-data packet in forward_packet()!\n");
    abort();
    break;
  }
finish_pkt:
  // pass along
  cmh->addr_type_ = AF_INET;
  cmh->xmit_failure_ = mac_callback;
  cmh->xmit_failure_data_ = this;
  // point the packet *down* the stack
  cmh->direction_ = -1;
  // data packet can serve as implicit beacon; put self in hops_[0]
  double myx, myy, myz;
  mn_->getLoc(&myx, &myy, &myz);
  /* the packet may *already* have hops stored; don't allocate with
     add_hop()! */
  gpsrh->hops_[0].ip = mn_->address();
  gpsrh->hops_[0].x = myx;
  gpsrh->hops_[0].y = myy;
  gpsrh->hops_[0].z = myz;

  if (verbose_)
    trace ("VFP %.5f _%d -> %d_ %d:%d -> %d:%d", now, mn_->address(),
	   ne->dst,
	   Address::instance().get_nodeaddr(iph->src_),
	   iph->sport_,
	   Address::instance().get_nodeaddr(iph->dst_),
	   iph->dport_);
  target_->recv(p, (Handler *)0);
  if (use_implicit_beacon_)
    BEACON_RESCHED;
}

void GPSR_Agent::recv(Packet *p, Handler *)
{
  hdr_ip *iph = (hdr_ip *) p->access(off_ip_);
  hdr_cmn *cmh = (hdr_cmn *) p->access(off_cmn_);
  hdr_gpsr *gpsrh = hdr_gpsr::access(p);
  MobileNode *dmn;

  int src = Address::instance().get_nodeaddr(iph->src_);
  int dst = Address::instance().get_nodeaddr(iph->dst_);

  if ((src == mn_->address()) &&
      (cmh->num_forwards() == 0)) {
    // Must be a packet I'm originating...
    // add the IP Header and dst position
    cmh->size() += IP_HDR_LEN + 8;
    // add bytes for hop-by-hop position/ip, if we're doing implicit beaconing
    if (use_implicit_beacon_)
      cmh->size() += 12;
    // iph->ttl_ = IP_DEF_TTL;
    iph->ttl_ = 128;
    // Fill in the destination's position
    dmn = ldb_->nlookup(dst);
    dmn->getLoc(&iph->dx(), &iph->dy(), &iph->dz());
    if (iph->dport_ != GPSR_PORT)
      // all data packets begin life in greedy mode
      gpsrh->mode_ = GPSRH_DATA_GREEDY;
  }
  // I received a packet that I sent.
  else if ((src == mn_->address()) && (gpsrh->mode_ == GPSRH_DATA_GREEDY)) {
    /* GPSR peri probes and peri data packets *can* visit the src twice;
       only greedy data packets shouldn't. */
    drop(p, DROP_RTR_ROUTE_LOOP);
    return;
  }
  else {
    /* packet I'm forwarding.
       Check the TTL. If it is zero, then discard. */
    if (--iph->ttl_ == 0) {
      drop(p, DROP_RTR_TTL);
      return;
    }
  }
  
  if (iph->dport_ == GPSR_PORT) {
    char *as;
    // callees here must free the packet if they don't forward it!!!
    switch (gpsrh->mode_) {
    case GPSRH_BEACON:
      if (src != mn_->address())
	// don't receive my own beacons; I'm not my neighbor
	beaconIn(p, gpsrh);
      break;
    case GPSRH_PPROBE:
      periIn(p, gpsrh);
      break;
    case GPSRH_DATA_GREEDY:
      as = Address::instance().print_nodeaddr(addr_);
      fprintf(stderr, "greedy data pkt @ %s:GPSR_PORT!\n",
	      as);
      fflush(stderr);
      delete[] as;
      break;
    case GPSRH_DATA_PERI:
      as = Address::instance().print_nodeaddr(addr_);
      fprintf(stderr, "peri data pkt @ %s:GPSR_PORT!\n",
	      as);
      fflush(stderr);
      delete[] as;
      break;
    default:
      as = Address::instance().print_nodeaddr(addr_);
      fprintf(stderr, "unk pkt type %d @ %s:GPSR_PORT!\n", gpsrh->mode_, as);
      fflush(stderr);
      delete[] as;
      break;
    }
  }
  else
    forwardPacket(p);
}

int GPSR_Agent::command(int argc, const char *const *argv)
{
  if (argc == 2) {
    if (strcmp(argv[1], "start-gpsr") == 0) {
      init();
      return TCL_OK;
    }
  }
  else if (argc == 3) {
    TclObject *obj;

    if (strcasecmp(argv[1], "tracetarget") == 0) {
      if ((obj = TclObject::lookup(argv[2])) == 0) {
	fprintf(stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1],
		argv[2]);
	return TCL_ERROR;
      }
      tracetarget = (Trace *) obj;
      return TCL_OK;
    }
    if (strcmp(argv[1], "install-tap") == 0) {
      if ((obj = TclObject::lookup(argv[2])) == 0) {
	fprintf(stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1],
		argv[2]);
	return TCL_ERROR;
      }
      Mac *m = (Mac *) obj;
      m->installTap(this);
      return TCL_OK;
    }
    if (strcmp(argv[1], "node") == 0) {
      if ((obj = TclObject::lookup(argv[2])) == 0) {
	fprintf(stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1],
		argv[2]);
	return TCL_ERROR;
      }
      mn_ = (MobileNode *) obj;
      return TCL_OK;
    }
    if (strcmp(argv[1], "ldb") == 0) {
      if ((obj = TclObject::lookup(argv[2])) == 0) {
	fprintf(stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1],
		argv[2]);
	return TCL_ERROR;
      }
      ldb_ = (LocDbase *) obj;
      return TCL_OK;
    }
    if (strcmp(argv[1], "add-ifq") == 0) {
      if ((obj = TclObject::lookup(argv[2])) == 0) {
	fprintf(stderr, "%s: %s lookup of %s failed\n", __FILE__, argv[1],
		argv[2]);
	return TCL_ERROR;
      }
      ifq_ = (PriQueue *) obj;
      return TCL_OK;
    }
  }
  return (Agent::command (argc, argv));
}

void GPSR_Agent::init(void)
{
  beacon_timer_ = new GPSR_BeaconTimer(this);
  beacon_timer_->sched(Random::uniform(bint_));
  if (!peri_proact_) {
    lastperi_timer_ = new GPSR_LastPeriTimer(this);
    lastperi_timer_->sched(lpexp_);
  }
  if (use_timed_plnrz_) {
    planar_timer_ = new GPSR_PlanarTimer(this);
    // XXX should make interval configurable!!
    planar_timer_->sched(1.0);
  }
}
