/*

  Copyright (C) 2000 President and Fellows of Harvard College

  All rights reserved.

  NOTICE: This software is provided "as is", without any warranty,
  including any implied warranty for merchantability or fitness for a
  particular purpose.  Under no circumstances shall Harvard University
  or its faculty, staff, students or agents be liable for any use of,
  misuse of, or inability to use this software, including incidental
  and consequential damages.

  License is hereby given to use, modify, and redistribute this
  software, in whole or in part, for any commercial or non-commercial
  purpose, provided that the user agrees to the terms of this
  copyright notice, including disclaimer of warranty, and provided
  that this copyright notice, including disclaimer of warranty, is
  preserved in the source code and documentation of anything derived
  from this software.  Any redistributor of this software or anything
  derived from this software assumes responsibility for ensuring that
  any parties to whom such a redistribution is made are fully aware of
  the terms of this license and disclaimer.

  Author: Brad Karp, Harvard University EECS, May, 1999

*/

// GPSR for ns2 w/wireless extensions

#include "agent.h"
#include "ip.h"
#include "delay.h"
#include "scheduler.h"
#include "queue.h"
#include "trace.h"
#include "arp.h"
#include "ll.h"
#include "mac.h"
#include "priqueue.h"
#include "node.h"
#include "timer-handler.h"

#define GPSR_PORT 0xff		// port number for GPSR protocol messages

#define GPSR_ALIVE_DESYNC 0.5	// desynchronizing term for alive beacons
#define GPSR_ALIVE_INT 0.5	// interval between alive beacons
#define GPSR_ALIVE_EXP (3*(GPSR_ALIVE_INT+GPSR_ALIVE_DESYNC*GPSR_ALIVE_INT))
				// timeout for expiring rx'd beacons

#define GPSR_PPROBE_INT 1.5	// interval between perimeter probes
#define GPSR_PPROBE_DESYNC 0.5	// desynchronizing term for perimeter probes
#define GPSR_PPROBE_EXP 8.0	// how often must use a perimeter to keep
				// probing them

#define GPSR_PPROBE_RTX 1

class NeighbEnt;
class NeighbTable;
class GPSR_BeaconTimer;
class GPSR_LastPeriTimer;
class GPSR_PlanarTimer;
class hdr_gpsr;

inline int cross_segment(double, double, double, double,
			 double, double, double, double,
			 double * = 0, double * = 0);

class GPSR_Agent : public Tap, public Agent {
public:
  GPSR_Agent(void);
  void beacon_callback(void);	// generate a beacon (timer-triggered)
  void deadneighb_callback(NeighbEnt *ne); // neighbor gone (timer/MAC-trig)
  void periprobe_callback(NeighbEnt *ne); // gen perimeter probe (timer-trig)
  void lastperi_callback(void);	// turn off peri probes when unused for timeout
  void planar_callback(void);	// planarization callback
  virtual int command(int argc, const char * const * argv);
  void lost_link(Packet *p);
  void tap(const Packet *p);

protected:
  NeighbTable *ntab_;
  GPSR_BeaconTimer *beacon_timer_; // Alive beacon timer
  GPSR_LastPeriTimer *lastperi_timer_; // Last perimeter used timer
  GPSR_PlanarTimer *planar_timer_; // inter-planarization timer
  Trace *tracetarget;		// Trace Target
  virtual void recv(Packet *, Handler *);
  Packet *makeAlive(void);
  void beaconIn(Packet *, hdr_gpsr *);
  void trace(char *fmt, ...);
  void tracepkt(Packet *, double, int, const char *);
  void forwardPacket(Packet *, int = 0);
  void periIn(Packet *, hdr_gpsr *, int = 0);
  int crosses(NeighbEnt *, hdr_gpsr *);
  void beacon_proc(int, double, double, double);
  void init();
  int use_mac_;			// whether or not to simulate full MAC level
  int use_peri_;		// whether or not to use perimeters
  int verbose_;			// verbosity (binary)
  int drop_debug_;		// whether or not to be verbose on NRTE events
  int peri_proact_;		// whether or not to pro-actively send pprobes
  int use_implicit_beacon_;	// whether or not all data packets are beacons
  int use_planar_;		// whether or not to planarize graph
  int use_loop_detect_;		// whether or not to fix loops in peridata pkts
  int use_timed_plnrz_;		// whether or not to replanarize w/timer
  double bint_;			// beacon interval
  double bdesync_;		// beacon desync random component range
  double bexp_;			// beacon expiration interval
  double pint_;			// perimeter probe interval
  double pdesync_;		// perimeter probe desync random cpt. range
  double lpexp_;		// perimeter probe generation timeout
  LocDbase *ldb_;		// location database
  MobileNode *mn_;		// my MobileNode
  PriQueue *ifq_;		// my ifq
  int off_gpsr_;		// offset of the GPSR packet header in pkt
};

// an entry in a perimeter header
struct PeriEnt {
  double x;
  double y;
  double z;
  nsaddr_t ip;
};

#define PERI_DEFAULT_HOPS 32	// default max number of hops in peri header

#define GPSRH_DATA_GREEDY 0	// greedy mode data packet
#define GPSRH_DATA_PERI 1	// perimeter mode data packet
#define GPSRH_PPROBE 2		// perimeter probe packet
#define GPSRH_BEACON 3		// liveness beacon packet

#ifdef HDR_GPSR_DYNAMIC
/* Maybe someday ns will support dynamic header allocation... */
class hdr_gpsr {
public:
  static int offset_;
  struct PeriEnt *hops_;
  struct PeriEnt peript_;
  nsaddr_t periptip_[3];
  int maxhops_;
  int nhops_;
  int currhop_;
  int mode_;
  hdr_gpsr() {
    hops_ = new struct PeriEnt[PERI_DEFAULT_HOPS];
    maxhops_ = PERI_DEFAULT_HOPS;
    nhops_ = 0;
  }
  ~hdr_gpsr() { delete[] hops_; }
  // header size is source route + int mode, int src rt len, int src rt ptr
  int size() { return (nhops_ * 12 + 3 * sizeof(int)); }
  static hdr_gpsr *access(Packet* p) {
    return (hdr_gpsr *) p->access(offset_);
  }
  void add_hop(nsaddr_t addip, double addx, double addy, double addz) {
    if (nhops_ == maxhops_) {
      // grow hops array
      struct PeriEnt *tmp;
      tmp = new struct PeriEnt[nhops_ << 1];
      bcopy(hops_, tmp, nhops_*sizeof(struct PeriEnt));
      delete[] hops_;
      hops_ = tmp;
      maxhops_ <<= 1;
    }
    hops_[nhops_].x = addx; hops_[nhops_].y = addy; hops_[nhops_].z = addz;
    hops_[nhops_].ip = addip;
    nhops_++;
  }
};
#else
/* ns bogusly doesn't invoke constructors for headers; it allocates them as
   raw byte fields. It also doesn't support changing the allocated size of
   a header on an in-flight packet. For now, give in and use pre-allocated
   headers with fixed maximum numbers of hosts. */
#define MAX_PERI_HOPS_STATIC 128
class hdr_gpsr {
public:
  static int offset_;
  struct PeriEnt hops_[MAX_PERI_HOPS_STATIC];
  struct PeriEnt peript_;
  nsaddr_t periptip_[3];
  int nhops_;
  int currhop_;
  int mode_;
  // header size is source route + int mode, int src rt len, int src rt ptr
  int size() { return (nhops_ * 12 + 3 * sizeof(int)); }
  static hdr_gpsr *access(Packet* p) {
    return (hdr_gpsr *) p->access(offset_);
  }
  void add_hop(nsaddr_t addip, double addx, double addy, double addz) {
    if (nhops_ == MAX_PERI_HOPS_STATIC) {
      fprintf(stderr, "hdr_gpsr::add_hop: out of slots!\n");
      abort();
    }
    hops_[nhops_].x = addx; hops_[nhops_].y = addy; hops_[nhops_].z = addz;
    hops_[nhops_].ip = addip;
    nhops_++;
  }
};  
#endif // HDR_GPSR_DYNAMIC

class GPSR_AgentTimer : public TimerHandler {
public:
  GPSR_AgentTimer(GPSR_Agent *a_) { a = a_; }
  virtual void expire(Event *e) = 0;

protected:
  GPSR_Agent *a;
};

class GPSR_NeighbTimer : public GPSR_AgentTimer {
public:
  GPSR_NeighbTimer(GPSR_Agent *a_, NeighbEnt *ne_) :
    GPSR_AgentTimer(a_) { ne = ne_; }
  virtual void expire(Event *e) = 0;

protected:
  NeighbEnt *ne;
};

class GPSR_BeaconTimer : public GPSR_AgentTimer {
public:
  GPSR_BeaconTimer(GPSR_Agent *a_) : GPSR_AgentTimer(a_) { };
  virtual void expire(Event *) { a->beacon_callback(); }
};

class GPSR_LastPeriTimer : public GPSR_AgentTimer {
public:
  GPSR_LastPeriTimer(GPSR_Agent *a_) : GPSR_AgentTimer(a_) { };
  virtual void expire(Event *) { a->lastperi_callback(); }
};

class GPSR_PlanarTimer : public GPSR_AgentTimer {
public:
  GPSR_PlanarTimer(GPSR_Agent *a_) : GPSR_AgentTimer(a_) { };
  virtual void expire(Event *) { a->planar_callback(); }
};

class GPSR_DeadNeighbTimer : public GPSR_NeighbTimer {
public:
  GPSR_DeadNeighbTimer(GPSR_Agent *a_, NeighbEnt *ne_) :
    GPSR_NeighbTimer(a_, ne_) { };
  virtual void expire(Event *) { a->deadneighb_callback(ne); }
};

class GPSR_PeriProbeTimer : public GPSR_NeighbTimer {
public:
  GPSR_PeriProbeTimer(GPSR_Agent *a_, NeighbEnt *ne_) :
    GPSR_NeighbTimer(a_, ne_) { };
  virtual void expire(Event *) { a->periprobe_callback(ne); }
};

#define PLANARIZE_RNG 0
#define PLANARIZE_GABRIEL 1

class NeighbEnt {
public:
  NeighbEnt(GPSR_Agent *ina) :
    peri(NULL), perilen(0), maxlen(0), dnt(ina, this), ppt(ina, this) { };
  void planarize(NeighbTable *, int,
		 double, double, double); // screen this edge
  int closer_pt(nsaddr_t myip, double myx, double myy, double myz,
		double ptx, double pty, nsaddr_t ptipa, nsaddr_t ptipb,
		double dstx, double dsty, double *closerx, double *closery);
  nsaddr_t dst;			// IP of neighbor
  double x, y, z;		// location of neighbor last heard
  struct PeriEnt *peri;		// perimeter via this neighbor
  int perilen;			// length of perimeter
  int maxlen;			// allocated slots in peri
  GPSR_DeadNeighbTimer dnt;	// timer for expiration of neighbor
  GPSR_PeriProbeTimer ppt;	// timer for generation of perimeter probe to
				//   neighbor
  int live;			// when planarizing, whether edge should be
				//   used
};

// opaque type: returned by NeighbTable iterator, holds place in table
typedef unsigned int NeighbTableIter;

// underlying data structure: an array, ordered by dst addr
class NeighbTable {
public:
  NeighbTable(GPSR_Agent *mya);
  ~NeighbTable();
  void ent_delete(const NeighbEnt *ent);
  NeighbEnt *ent_add(const NeighbEnt *ent);
  NeighbEnt *ent_finddst(nsaddr_t dst);
  NeighbEnt *ent_findshortest(MobileNode *mn, double x, double y,
			      double z);
  NeighbEnt *ent_findcloser_onperi(MobileNode *mn, double x, double y,
				   double z, int *perihop);
  NeighbEnt *ent_findcloser_edgept(MobileNode *, double, double, nsaddr_t,
				   nsaddr_t, double, double, double *,
				   double *);
  NeighbEnt *ent_next_ccw(MobileNode *, NeighbEnt *, int);
  NeighbEnt *ent_next_ccw(double, double, double, int, NeighbEnt * = 0);
  NeighbEnt *ent_findface(MobileNode *, double, double, double, int);
  NeighbTableIter InitLoop();	// iterator initializer
  NeighbEnt *NextLoop(NeighbTableIter *); // iterator
  void planarize(int, double, double, double); // remove all crossing edges
private:
  int nents;			// current entries in use
  int maxents;			// allocated table size
  GPSR_Agent *a;		// agent that owns this table
  NeighbEnt **tab;
};
