/* mac-null.cc
   $Id: mac-null.cc,v 1.3 2000/02/20 01:16:46 bkarp Exp $

   implement a mac layer that IGNORES collisions.  This layer allows the 
   simultaneous receipt of multiple packets.   

   If the variable no_delay_send_ is set to true
     The layer delays packet 1 tx time on receipt, but sends back to back 
     packets immeadiately one after the other.  Together, the effect is as
     if the link had infinite BW, and the txtime is reassigned to propagation
     delay.
   else, the txtime is used to pace the sending of packets
   
   */

#include "delay.h"
#include "connector.h"
#include "packet.h"
#include "random.h"

// #define DEBUG

#include "arp.h"
#include "ll.h" 
#include "mac.h"
#include "mac-802_11.h"
#include "cmu-trace.h"

// stolen from mac-802_11.h since I don't have ml's cmu-trace that groks
// other than 802.11 headers -dam 4/12/99
struct hdr_macnull {
	struct frame_control	dh_fc;
	u_int16_t		dh_duration;
	u_char			dh_da[ETHER_ADDR_LEN];
	u_char			dh_sa[ETHER_ADDR_LEN];
	u_char			dh_bssid[ETHER_ADDR_LEN];
	u_int16_t		dh_scontrol;
	u_char			dh_body[0];
};


class MacNULL;

class MacNULLTimer : public Handler {
public:
	MacNULLTimer(MacNULL* m) :  mac_(m), busy_(0) {}
	virtual void handle(Event *e) = 0; 
	virtual inline void cancel() {
		assert(0);
	}
	virtual void schedule(double t, Packet *p);
	inline int busy(void) { return busy_; }
	inline double expire(void) { return intr.time_; }
protected:
	MacNULL	        *mac_;
	Event		intr;
	int		busy_;
};

class MacNULLSend : public MacNULLTimer {
public:
	MacNULLSend(MacNULL* m) :  MacNULLTimer(m) {}
	virtual void handle(Event *e); 
};

class MacNULLRecv : public MacNULLTimer {
public:
	MacNULLRecv(MacNULL* m) :  MacNULLTimer(m) {}
	virtual void handle(Event *e); 
	virtual void schedule(double t, Packet *p);
};

class MacNULL : public Mac {
	friend MacNULLSend;
	friend MacNULLRecv;
public:
	MacNULL();
	void		recv(Packet *p, Handler *h);
	inline int	hdr_dst(char* hdr, int dst = -2);
        inline int      hdr_dst(hdr_macnull *hdr, int dst = -2) {
          return (MacNULL::hdr_dst((char *)hdr, dst));
	}
	inline int	hdr_src(char* hdr, int src = -2);
        inline int      hdr_src(hdr_macnull *hdr, int src = -2) {
	  return (MacNULL::hdr_src((char *)hdr, src));
	}
	inline int	hdr_type(char* hdr, u_int16_t type = 0);

private:
	int		command(int argc, const char*const* argv);
	void		discard(Packet *p, const char* why);
	void	        send(Packet *p, Handler *h);

	MacNULLSend     send_timer_;
	MacNULLRecv     recv_timer_;
	void            sendComplete(Packet *p);
	void            recvComplete(Packet *p);
	Handler         *upcall_;
	int             no_delay_send_; // if true, packets can be sent back-to-back

	/*
	 * Debugging Functions.
	 */
	void		trace_pkt(Packet *p);
	void		dump(char* fname);

	inline int initialized() {
		return (logtarget_ && Mac::initialized());
	}

	void mac_log(Packet *p) {
		logtarget_->recv(p, (Handler*) 0);
	}

private:
	int off_arp_;

	/* ============================================================
	   Internal MAC State
	   ============================================================ */

	int		min_frame_len;
	NsObject*	logtarget_;

	/* ============================================================
	   Duplicate Detection state
	   ============================================================ */
	u_int16_t	sta_seqno;	// next seqno that I'll use
	int		cache_node_count;
};
	
static class MacNULLClass : public TclClass {
public:
	MacNULLClass() : TclClass("Mac/NULL") {}
	TclObject* create(int, const char*const*) {
		return (new MacNULL);
	}
} class_macNULL;

/* ======================================================================
   Mac Class Functions
   ====================================================================== */
MacNULL::MacNULL(): send_timer_(this), recv_timer_(this), 
	upcall_(0), no_delay_send_(0)
{
	sta_seqno = 1;
	cache_node_count = 0;

	bind("off_arp_", &off_arp_);
	//bind("no_delay_send_", &no_delay_send_);

	fprintf(stderr,"WARNING: null MAC layer selected.  ignoring collisions\n");	
}


/* ======================================================================
   Packet Headers Routines
   ====================================================================== */
	
inline int
MacNULL::hdr_dst(char* hdr, int dst)
{
	struct hdr_macnull *mh = (struct hdr_macnull*) hdr;

	if(dst > -2)
		STORE4BYTE(&dst,mh->dh_da);

	return ETHER_ADDR(mh->dh_da);
}

inline int
MacNULL::hdr_src(char* hdr, int src)
{
	struct hdr_macnull *mh = (struct hdr_macnull*) hdr;

	if(src > -2)
		STORE4BYTE(&src,mh->dh_sa);

	return ETHER_ADDR(mh->dh_sa);
}

inline int 
MacNULL::hdr_type(char* hdr, u_int16_t type)
{
	struct hdr_macnull *mh = (struct hdr_macnull*) hdr;

	if(type)
	        SET_ETHER_TYPE(mh->dh_body, type);

	return GET_ETHER_TYPE(mh->dh_body);
}


int
MacNULL::command(int argc, const char*const* argv)
{
	if (argc == 3) {

		if (strcmp(argv[1], "log-target") == 0) {
			logtarget_ = (NsObject*) TclObject::lookup(argv[2]);
			if(logtarget_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if(strcmp(argv[1], "nodes") == 0) {
			cache_node_count = atoi(argv[2]);
			return TCL_OK;
			
		}
	}
	return Mac::command(argc, argv);
}


/* ======================================================================
   Debugging Routines
   ====================================================================== */
void
MacNULL::trace_pkt(Packet *p) {
	struct hdr_cmn *ch = HDR_CMN(p);

	fprintf(stderr, "\t_%d_ %s %d\n",		
		index_, packet_info.name(ch->ptype()), ch->size());
}

void
MacNULL::dump(char *fname)
{
	fprintf(stderr,
		"\n%s --- (INDEX: %d, time: %2.9f)\n",
		fname, index_, Scheduler::instance().clock());
}



/* ======================================================================
   Incoming Packet Routines
   ====================================================================== */

void
MacNULL::recv(Packet *p, Handler *h)
{
	struct hdr_cmn *hdr = HDR_CMN(p);
	hdr_macnull* mh = (hdr_macnull*)p->access(off_mac_);

	/*
	 * Sanity Check
	 */
	assert(initialized());

	/*
	 *  Handle outgoing packets.
	 */
	if(hdr->direction() == -1) {
		// give target a copy, so we can still use p
                downtarget_->recv(p->copy(), h);

		if (no_delay_send_) {
			// tell IFQ to send down the next packet immeadiately
			// NOTE:  see comments at top of file
			upcall_ = h;
			sendComplete(p);
		} else {
			assert(upcall_ == 0);
			upcall_ = h;
			send_timer_.schedule(netif_->txtime(p),p);
		}
		return;
        }

	/*
	 *  Handle incoming packets.
	 *
	 *  We just received the 1st bit of a packet on the network
	 *  interface.
	 *
	 */

	if ((u_int32_t) hdr_dst(mh) !=  (u_int32_t) index_ &&
	    (u_int32_t) hdr_dst(mh) !=  MAC_BROADCAST && !tap_) {
		/*
		 *  We don't want to log this event, so we just free
		 *  the packet instead of calling the drop routine.
		 */
		Packet::free(p);
		goto done;
	}

	if (hdr->error() ) {
		discard(p, DROP_MAC_PACKET_ERROR);
		goto done;
	}

	/*
	 * reception of this packet will finish in
	 * txtime seconds.  schedule completion then
	 */
	recv_timer_.schedule(netif_->txtime(p), p);

done:
	return;
}


void 
MacNULL::discard(Packet *p, const char* why)
{
	hdr_macnull * mh = (hdr_macnull *)p->access(off_mac_);

	if ((u_int32_t) hdr_dst(mh) == (u_int32_t) index_)
		drop(p, why);
	else
		Packet::free(p);
}

void
MacNULL::send(Packet *, Handler *)
{
	fprintf(stderr,"DFU: shouldn't ever reach here %s\n",
		__PRETTY_FUNCTION__);
	abort();
}

void
MacNULL::sendComplete(Packet *p)
{
	hdr_macnull* mh = (hdr_macnull*)p->access(off_mac_);

	// provide a mac layer failure indication if needed.
	// this is completely bogus, but this is a test-only MAC
	// layer anyway
	if ((u_int32_t) hdr_dst(mh) != MAC_BROADCAST) {
		if (hdr_src(mh) != hdr_dst(mh)
		    && God::instance()->hops(hdr_src(mh), hdr_dst(mh)) == 0) {
			fprintf(stderr,"Mac/NULL mac layer requires that scenario file have\n"
				"  god data added.  see calcdest or setdest\n");
			abort();
		}
		if (God::instance()->hops(hdr_src(mh), hdr_dst(mh)) > 1) {
			/* tell the callback the send operation failed  */
			hdr_cmn *ch = HDR_CMN(p);
			if (ch->xmit_failure_) {
				ch->xmit_reason_ = XMIT_REASON_RTS;
				ch->xmit_failure_(p->copy(), ch->xmit_failure_data_);
			}
		}
	}
	
	Packet::free(p); // no drop reason needed

	// have the upper layer (probably the IFQ) unblock itself now
	assert(upcall_ != 0);
	Handler *h = upcall_;
	upcall_ = 0;
	h->handle((Event *) 0);
}

void
MacNULL::recvComplete(Packet *p)
{
	hdr_cmn *ch = HDR_CMN(p);
	hdr_macnull* mh = (hdr_macnull*)p->access(off_mac_);

	ch->num_forwards()++;

	if (tap_) {
		tap_->tap(p);
	}

	if ((u_int32_t) hdr_dst(mh) ==  (u_int32_t) index_ ||
	    (u_int32_t) hdr_dst(mh) ==  MAC_BROADCAST)
	  uptarget_->recv(p, (Handler *) 0);
	else
	  Packet::free(p);
}

void
MacNULLTimer::schedule(double t, Packet *p)
{
	Scheduler& s = Scheduler::instance();
	assert(busy_ == 0);
	s.schedule(this, p, t);
	busy_ = 1;
}

void 
MacNULLSend::handle(Event *e)
{
	Packet *p = (Packet *)e;  // cast of trust
	busy_ = 0;
	mac_->sendComplete(p);
}

void
MacNULLRecv::schedule(double t, Packet *p)
{
	Scheduler& s = Scheduler::instance();
	busy_++;
	s.schedule(this, p, t);
}

void 
MacNULLRecv::handle(Event *e)
{
	Packet *p = (Packet *)e;  // cast of trust
	busy_--;
	mac_->recvComplete(p);
}
